// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.Dimension;

/** Represents a point mass which forces can be applied to. */
public class Particle
{
	/** Mass of particle. */
	private float mass;
	/** Position of particle. */
	private Vector2d position;
	/** Last position of particle. */
	private Vector2d lastPosition;
	/** Velocity of particle. */
	private Vector2d velocity = new Vector2d(0, 0);
	/** Net force acting on particle. */
	private Vector2d netForce = new Vector2d(0, 0);
	/** Indicates if particle can move. */
	private boolean fixed = false;

	/** Constructs a new particle.
	 *  @param  initPosition             particle position.
	 *  @param  initMass                 particle mass.
	 *  @throws IllegalArgumentException if mass is negative.
	 */
	public Particle(Vector2d initPosition, float initMass)
	{
		// If passed negative mass
		if (initMass < 0)
		{
			throw new IllegalArgumentException("Partciles cannot have negative mass.");
		}
		
		position     = new Vector2d(initPosition);
		lastPosition = new Vector2d(initPosition);
		mass         = initMass;
	}

	/** Applies any forces acting on particle to velocity and updates its position.
	 *  @param  deltaTime amount of time to advance in seconds.
	 */
	public void update(float deltaTime)
	{
		// If particle is not fixed and has mass
		if (fixed == false && mass > 0)
		{
			// Calculate acceleration
			Vector2d acceleration = netForce.divide(mass);
			// Add acceleration to velocity
			velocity = velocity.add(acceleration.multiply(deltaTime));
		}

		// Reset net force
		netForce = new Vector2d(0, 0);
		// Store last position
		lastPosition = new Vector2d(position);
		// Calculate new position
		position = position.add(velocity.multiply(deltaTime));
	}

	/** Applies force to the particle.
	 *  @param force force to apply.
	 */
	public void applyForce(Vector2d force)
	{
		// Add force to net force
		netForce = netForce.add(force);
	}

	/** Applies viscous damping force.
	 *  @param drag drag coeffecient.
	 */
	public void applyDamping(float drag)
	{
		// Apply damping force
		applyForce(velocity.multiply(-drag));
	}

	/** Applies collision forces if particle's position in going outside set boundaries. Particle position is also backed up to prevent interpentration.
	 *  @param frictionOn               whether to apply friction or not if collision occurs.
	 *  @param bounds                   boundary to check collision against.
	 *  @param coefficientOfRestitution coefficent of restitution to apply if collision occurs.
	 *  @param staticSpeed              speed where particle is considered not moving.
	 *  @param staticFriction           coeffecient of static friction.
	 *  @param kineticFriction          coeffecient of kinetic friction.
	 */
	public void applyCollisions(boolean frictionOn, Dimension bounds, float coefficientOfRestitution, float staticSpeed, float staticFriction, float kineticFriction)
	{
		// Collision contact normal
		Vector2d contactNormal = new Vector2d(0, 0);
		// Whether a collision has occured
		boolean collision = false;		

		// Particle is left of boundary
		if (position.x < 0)
		{
			// Collision has occured
			collision = true;
			// Add collision normal
			contactNormal = contactNormal.add(new Vector2d(+1, 0));
			// Clip horizontal position
			position.x = 0;
		}

		// Particle is right of boundary
		if (position.x > bounds.width)
		{
			// Collision has occured
			collision = true;
			// Add collision normal
			contactNormal = contactNormal.add(new Vector2d(-1, 0));
			// Clip horizontal position
			position.x = bounds.width;
		}

		// Particle is above boundary
		if (position.y < 0)
		{
			// Collision has occured
			collision = true;
			// Add collision normal
			contactNormal = contactNormal.add(new Vector2d(0, +1));
			// Clip vertical position
			position.y = 0;
		}

		// Particle is below boundary
		if (position.y > bounds.height)
		{
			// Collision has occured
			collision = true;
			// Add collision normal
			contactNormal = contactNormal.add(new Vector2d(0, -1));
			// Clip vertical position
			position.y = bounds.height;
		}

		// Stop if no collisions occured
		if (!collision)
			return;
		
		// Normalise contact normal
		contactNormal = contactNormal.normalise();

		// Calculate velocity in contact normal direction
		Vector2d normalVelocity  = contactNormal.multiply(contactNormal.dot(velocity));
		// Calculate velocity perpendicular to contact normal direction
		Vector2d tangentVelocity = velocity.subtract(normalVelocity);

		// If friction is on
		if (frictionOn)
		{
			// Get size of tangent velocity
			float tangentVelocityMagnitude = tangentVelocity.getLength();
			// Calculate term in friction force
			Vector2d term = tangentVelocity.normalise().multiply(contactNormal.dot(netForce));

			// If particle is not moving
			if (tangentVelocityMagnitude > staticSpeed)
			{
				// Apply kinetic friction
				applyForce(term.multiply(kineticFriction));
			}
			else
			{
				// Apply static friction
				applyForce(term.multiply(staticFriction * (tangentVelocityMagnitude / staticSpeed)));
			}
		}

		// Calculate velocity after collision
		velocity = tangentVelocity.subtract(normalVelocity.multiply(coefficientOfRestitution));
	}

	/** Sets particle's velocity and forces acting on it to zero. */
	public void stop()
	{
		// Set velocity and net force to zero to stop movement
		velocity = new Vector2d(0, 0);
		netForce = new Vector2d(0, 0);
	}

	/** Sets fixed status of particle.
	 *  @param newFixed new fixed status.
	 */
	public void setFixed(boolean newFixed)
	{
		fixed = newFixed;

		// If particle being fixed
		if (fixed)
		{
			// Stop it moving
			stop();
		}	
	}

	/** Sets position of particle.
	 *  @param newPostion new position.
	 */
	public void setPosition(Vector2d newPosition)
	{
		position = new Vector2d(newPosition);
	}

	/** Sets velocity of particle.
	 *  @param newVelocity new velocity.
	 */
	public void setVelocity(Vector2d newVelocity)
	{
		velocity = new Vector2d(newVelocity);
	}

	/** Returns position of particle. */
	public Vector2d getPosition()
	{
		return new Vector2d(position);
	}

	/** Returns velocity of particle. */
	public Vector2d getVelocity()
	{
		return new Vector2d(velocity);
	}

	/** Returns fixed status of particle. */
	public boolean isFixed()
	{
		return fixed;
	}

	/** Returns mass of particle. */
	public float getMass()
	{
		return mass;
	}
}