// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.applet.*;
import java.awt.*;
import java.awt.event.*;

public class SandboxApplet extends Applet implements Runnable
{
	/** Play mode ID. */
	private static final int MODE_PLAY = 0;
	/** Build mode ID. */
	private static final int MODE_BUILD = 1;
	/** Remove mode ID. */
	private static final int MODE_REMOVE = 2;
	/** Glue mode ID. */
	private static final int MODE_GLUE = 3;

	/** Number of sub-steps to take per step, to increase simulation accuracy. */
	private static final int NUM_SUBSTEPS = 4;
	/** Length of time to update simulation each step. */
	private static final float STEP_SIZE = 0.01f;
	/** Maximum time between last mouse drag and button release to cause particle to be thrown. */
	private static final int THROW_TIME_THRESHOLD = 100;
	/** Amount to scale mouse displacement by when throwing. */
	private static final int THROW_AMPLIFICATION = 100;

	/** Default mass of particle. */
	private static final float PARTICLE_MASS = 1.00f;
	/** Default world damping. */
	private static final float WORLD_DAMPING = 0.05f;
	/** Default spring damping. */
	private static final float SPRING_DAMPING = 20.00f;
	/** Default spring stiffness. */
	private static final float SPRING_STIFFNESS = 350.00f;
	/** Velocity when particle is considered not moving. */
	private static final float STATIC_VELOCITY = 0.01f;
	/** Default static friction. */
	private static final float STATIC_FRICTION = 0.50f;
	/** Default kinetic friction. */
	private static final float KINETIC_FRICTION	= 0.30f;
	/** Default coeffecient of restitution. */
	private static final float COEFFICIENT_OF_RESTITUTION = 0.80f;
	/** Default gravitation pull vector. */
	private static final Vector2d GRAVITATIONAL_PULL = new Vector2d(0, +800);

	/** The simulation world. */
	private World world;
	/** Current mode. */
	private int mode = MODE_BUILD;

	/** Last mouse position in world view. */
	private Vector2d mousePosition;
	/** Particle being held by mouse. */
	private int grabbedParticle = World.NONE;
	/** Particle selected by mouse. */
	private int selectedParticle = World.NONE;
	/** Time last mouse drag event was logged in world view. */
	private long lastDragged = 0;
	/** Mouse displacement between last drag positions. */
	private Vector2d deltaPosition = new Vector2d(0, 0);
	/** Fixed status before a particle was grabbed. */
	private boolean wasFixed  = false;

	/** Control panel for manipulating world. */
	private ControlPanel controlPanel = new ControlPanel();
	/** Component for viewing world. */
	private WorldView worldView = new WorldView();

	/** Thread for updating simulation. */
	private Thread updateThread;

	public void init()
	{
		// Use border layout
		setLayout(new BorderLayout());
		// Add world view in center
		add(worldView,  BorderLayout.CENTER);
		// Add control panel at the bottom
		add(controlPanel, BorderLayout.SOUTH);
	}

	public void resetWorld()
	{
		// Create world
		world = new World(new Dimension(worldView.getSize().width, worldView.getSize().height), WORLD_DAMPING, COEFFICIENT_OF_RESTITUTION, STATIC_VELOCITY, STATIC_FRICTION, KINETIC_FRICTION, GRAVITATIONAL_PULL);
	}

	public void start()
	{
		// If world has not been created
		if (world == null)
		{
			// Create world
			resetWorld();
		}

		// Create update thread
		updateThread = new Thread(this);
		// Start update thread
		updateThread.start();
	}

	public void stop()
	{
		// Stop update thread
		updateThread = null;
	}

	public void destroy()
	{
		// Stop update thread
		updateThread = null;
	}

	public void run()
	{
        	// While this is the executing thread
		while (Thread.currentThread() == updateThread)
		{
			// If in play mode
			if (mode == MODE_PLAY)
			{
				// For each substep
				for (int i = 0; i < NUM_SUBSTEPS; ++i)
				{
					synchronized (world)
					{
						// Update world 
						world.update(STEP_SIZE / NUM_SUBSTEPS);
					}
				}
			}

			// Repaint world view
			worldView.repaint();

			try
			{
				Thread.sleep(10);
			}
			catch (InterruptedException e)
			{
			}
		}
	}

	public void update(Graphics g)
	{
		// Paint components
		paintAll(g);
	}

	public String getAppletInfo()
	{
		return "Sandbox - Copyright (c) 2001 Sean Wilson. All Rights Reserved.";
	}

	private class ControlPanel extends Panel
	{
		/** Play mode choice string. */
		private final String LABEL_PLAY    = "Play";
		/** Build mode choice string. */
		private final String LABEL_BUILD   = "Build";
		/** Remove mode choice string. */
		private final String LABEL_REMOVE  = "Remove";
		/** Glue mode choice string. */
		private final String LABEL_GLUE    = "Glue";
		/** Restart choice string. */
		private final String LABEL_RESTART = "Restart";

		public ControlPanel()
		{
			// Set background of component
			setBackground(Color.lightGray);
			
			// Mode choice component
			final Choice choice = new Choice();
		
			// Add choice to component
			add(choice);

			// Add labels to choice
			choice.addItem(LABEL_PLAY);
			choice.addItem(LABEL_BUILD);
			choice.addItem(LABEL_GLUE);
			choice.addItem(LABEL_REMOVE);
			choice.addItem(LABEL_RESTART);

			// Select build choice
			choice.select(LABEL_BUILD);
			// Set mode to build mode
			mode = MODE_BUILD;

			// Add a listener to choice
			choice.addItemListener(
				new ItemListener()
				{
					public void itemStateChanged(ItemEvent e)
					{
						// Get source of event
						Choice source = (Choice)e.getSource();
						
						// If play choice selected
						if (source.getSelectedItem() == LABEL_PLAY)
						{
							// Change to play mode
							mode = MODE_PLAY;
						}
						else if (source.getSelectedItem() == LABEL_BUILD) // If build choice selected
						{
							// Change to build mode
							mode = MODE_BUILD;
						}
						else if (source.getSelectedItem() == LABEL_REMOVE) // If remove choice selected
						{
							// Change to remove mode
							mode = MODE_REMOVE;
						}
						else if (source.getSelectedItem() == LABEL_GLUE) // If glue choice selected
						{
							// Change to glue mode
							mode = MODE_GLUE;
						}
						else if (source.getSelectedItem() == LABEL_RESTART) // If restart choice selected
						{
							// Reset world
							resetWorld();
							// Select build choice
							choice.select(LABEL_BUILD);
							// Change to build mode
							mode = MODE_BUILD;
						}

						// No selected particle
						selectedParticle = World.NONE;
					}
				}
			);
		}

		public void paint(Graphics g)
		{
			// Draw outline around component, except for top border
			g.drawRect(0, -1, getSize().width - 1, getSize().height - 1);	
		}
	}

	private class WorldView extends Component implements MouseListener, MouseMotionListener
	{
		/** Radius of particle circle. */
		private static final int PARTICLE_RADIUS = 8;
		/** Radius of particle highlight circle. */
		private static final int HIGHLIGHT_RADIUS = 16;
		/** Radius of circle particle can be selected in. */
		private static final int PARTICLE_SELECT_RADIUS = 20;
		/** Double buffer. */
		private Image offscreen;

		public WorldView()
		{
			// Add mouse listener and motion listener to component
			addMouseListener(this);
			addMouseMotionListener(this);
		}

		public void paintOutline(Graphics g)
		{
			// Draw black outline around component
			g.setColor(Color.black);
			g.drawRect(0, 0, getSize().width - 1, getSize().height - 1);
		}

		public void paintBackground(Graphics g)
		{
			// Fill background white
			g.setColor(Color.white);
			g.fillRect(0, 0, getSize().width, getSize().height);
		}

		public void paintParticle(Graphics g, Vector2d position, boolean fixed)
		{
			// If particle is fixed
			if (fixed)
			{
				// Draw white particle with black outline
				g.setColor(Color.white);
				g.fillOval((int)position.x - PARTICLE_RADIUS / 2, (int)position.y - PARTICLE_RADIUS / 2, PARTICLE_RADIUS, PARTICLE_RADIUS);
				g.setColor(Color.black);
				g.drawOval((int)position.x - PARTICLE_RADIUS / 2, (int)position.y - PARTICLE_RADIUS / 2, PARTICLE_RADIUS, PARTICLE_RADIUS);				
			}
			else
			{
				// Draw black particle
				g.setColor(Color.black);
				g.fillOval((int)position.x - PARTICLE_RADIUS / 2, (int)position.y - PARTICLE_RADIUS / 2, PARTICLE_RADIUS, PARTICLE_RADIUS);
			}			
		}

		public void paintParticleHighlight(Graphics g, Vector2d position)
		{
			// Draw particle highlight
			g.setColor(Color.black);
			g.drawOval((int)position.x - HIGHLIGHT_RADIUS / 2, (int)position.y - HIGHLIGHT_RADIUS / 2, HIGHLIGHT_RADIUS, HIGHLIGHT_RADIUS);
		}

		public void paintSpring(Graphics g, Vector2d position1, Vector2d position2)
		{
			// Draw spring
			g.setColor(Color.gray);
			g.drawLine((int)position1.x, (int)position1.y, (int)position2.x, (int)position2.y);
		}

		public void paintSprings(Graphics g)
		{
			// For each spring
			for (int i = 0; i < world.getNumSprings(); ++i)
			{
				// Paint spring
				paintSpring(g, world.getSpring(i).getParticle1Position(), world.getSpring(i).getParticle2Position());
			}

			// If mouse in component
			if (mousePosition != null)
			{
				// If no grabbed particle and a selected one
				if (grabbedParticle == World.NONE && selectedParticle != World.NONE)
				{
					// Paint a spring from selected particle to mouse
					paintSpring(g, mousePosition, world.getParticle(selectedParticle).getPosition());
				}
			}
		}

		public void paintParticles(Graphics g)
		{
			// For each particle
			for (int i = 0; i < world.getNumParticles(); ++i)
			{
				// Ignore if particle is fixed because it is grabbed
				boolean fixed = (grabbedParticle == i ? wasFixed : world.getParticle(i).isFixed());
				// Paint particle
				paintParticle(g, world.getParticle(i).getPosition(), fixed);
			}

			// If mouse is in component
			if (mousePosition != null)
			{
				// If particle is grabbed
				if (grabbedParticle != World.NONE)
				{
					// Highlight grabbed particle
					paintParticleHighlight(g, world.getParticle(grabbedParticle).getPosition());
				}
				else
				{
					// Get pointed at particle
					int pointedParticle = world.getParticle(mousePosition, PARTICLE_SELECT_RADIUS);

					// If a particle is being pointed at
					if (pointedParticle != World.NONE)
					{
						// Highlight pointed at particle
						paintParticleHighlight(g, world.getParticle(pointedParticle).getPosition());
					}
				}
			}
		}

		public void update(Graphics g)
		{
			// Paint component
			paint(g);
		}

		public void paint(Graphics g)
		{
			// If double buffer has not been created
			if (offscreen == null)
			{
				// Create double buffer
				offscreen = createImage(getSize().width, getSize().height);
			}
			
			// Get graphics context to double buffer
			Graphics offscreenGraphics = offscreen.getGraphics();
        
			// Paint background, springs, particles and outline to double buffer
			paintBackground(offscreenGraphics);
			paintSprings   (offscreenGraphics);
			paintParticles (offscreenGraphics);
			paintOutline   (offscreenGraphics);

			// Paint double buffer
			g.drawImage(offscreen, 0, 0, this);
		}

		public void mousePressed(MouseEvent e)
		{
			// Store mouse position
			mousePosition = new Vector2d(e.getX(), e.getY());
			// Get particle mouse is pointing at
			int pointedParticle = world.getParticle(mousePosition, PARTICLE_SELECT_RADIUS);

			// If mouse is pointing at particle
			if (pointedParticle != World.NONE)
			{
				// If in remove mode
				if (mode == MODE_REMOVE)
				{
					// Remove particle
					world.removeParticle(pointedParticle);
				}
				else
				{
					// Get poainted at particle
					Particle particle = world.getParticle(pointedParticle);
					// Make pointed particle grabbed particle
					grabbedParticle = pointedParticle;

					// Store particle's original fixed state
					wasFixed = particle.isFixed();
					// Fix particle
					particle.setFixed(true);
					
					// If in glue mode
					if (mode == MODE_GLUE)
					{
						// Flip particle's original fixed state
						wasFixed = !wasFixed;
					}
					else
					{
						// If in play mode
						if (mode == MODE_PLAY)
						{
							// Move particle to mouse position
							world.getParticle(grabbedParticle).setPosition(mousePosition);
						}
						else
						{
							// Selected particle being pointed at
							if (selectedParticle == pointedParticle)
							{
								// No particles selected
								selectedParticle = World.NONE;
							}
							else
							{
								// If particle selected and it is not attached to pointed to particle already
								if (selectedParticle != World.NONE && !world.isSpringAttached(selectedParticle, pointedParticle))
								{
									// Add spring between pointed at and selected particle
									world.addSpring(new Spring(world.getParticle(selectedParticle), particle, Spring.AT_REST_LENGTH, SPRING_STIFFNESS, SPRING_DAMPING));
								}

								// Selected particle becomes pointed at particle
								selectedParticle = pointedParticle;
							}
						}
					}
				}
			}
			else if (mode == MODE_BUILD) // If in build mode
			{
				// Add particle to world
				int addedParticle = world.addParticle(new Particle(mousePosition, PARTICLE_MASS));

				// If a particle was selected
				if (selectedParticle != World.NONE)
				{
					// Add spring between added particle and seleceted particle
					world.addSpring(new Spring(world.getParticle(selectedParticle), world.getParticle(addedParticle), Spring.AT_REST_LENGTH, SPRING_STIFFNESS, SPRING_DAMPING));
				}

				// Select added particle
				selectedParticle = addedParticle;
				// Grab added particle
				grabbedParticle = addedParticle;
				// Added particle was not fixed
				wasFixed = false;
			}
		}

		public void mouseDragged(MouseEvent e)
		{
			// Store last mouse position
			Vector2d lastMousePosition = mousePosition;
			// Store current mouse position
			mousePosition = new Vector2d(e.getX(), e.getY()).clip(getSize());
			// Store when this drag message was sent
			lastDragged = e.getWhen();
			
			// If particle grabbed
			if (grabbedParticle != World.NONE)
			{
				// If in build mode
				if (mode == MODE_BUILD)
				{
					// Select grabbed particle
					selectedParticle = grabbedParticle;
				}
				else if (mode == MODE_PLAY) // If in play mode
				{
					// If mouse was in component before and is now
					if (mousePosition != null && getBounds().contains(e.getX(), e.getY()))
					{
						// Calculate mouse displacement from last time it was stored
						deltaPosition = new Vector2d(lastMousePosition, mousePosition);
					}
					else
					{
						// No mouse displacement
						deltaPosition = new Vector2d(0, 0);
					}
				}

				// Move grabbed particle to mouse
				world.getParticle(grabbedParticle).setPosition(mousePosition);
			}
		}

		public void mouseMoved(MouseEvent e)
		{
			// Store mouse position
			mousePosition = new Vector2d(e.getX(), e.getY());
		}

		public void mouseReleased(MouseEvent e)
		{
			// If event not inside view component
			if (!getBounds().contains(e.getX(), e.getY()))
			{
				// Mouse position not in view component
				mousePosition = null;
			}
			
			// If particle grabbed
			if (grabbedParticle != World.NONE)
			{
				// Reset particle's fixed status
				world.getParticle(grabbedParticle).setFixed(wasFixed);
				
				// If in play mode, grabbed particle was not originally fixed and last dragged mouse position happened recently
				if (mode == MODE_PLAY && !wasFixed && Math.abs(e.getWhen() - lastDragged) < THROW_TIME_THRESHOLD)
				{
					// Set velocity of particle in direction of mouse movement
					world.getParticle(grabbedParticle).setVelocity(deltaPosition.multiply(THROW_AMPLIFICATION));
				}

				// Let go off particle
				grabbedParticle = World.NONE;
			}
		}

		public void mouseExited(MouseEvent e)
		{
			// Reset cursor to default
			setCursor(new Cursor(Cursor.DEFAULT_CURSOR));

			// Mouse position not in view component
			mousePosition = null;
		}

		public void mouseEntered(MouseEvent e)
		{
			// Set cursor to cross-hair so user can select more accurately
			setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
		}

		public void mouseClicked(MouseEvent e)
		{

		}
	}
}