// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

/** Represents a spring which can be attached to two particles and apply forces to each. */
public class Spring
{
	/** Represents spring is already at rest length. */
	static final int AT_REST_LENGTH =  -1;
	
	/** First particle spring is attached to. */
	private Particle particle1;
	/** Second particle spring is attached to. */
	private Particle particle2;
	/** Spring damping constant. */
	private float damping;
	/** Spring stiffness/Hooke's constant. */
	private float stiffness;
	/** Spring rest length. */
	private float restLength;

	/** Constructs a new spring. Spring must be attached to two particles.
	 *  @param initParticle1  first particle spring is attached to.
	 *  @param initParticle2  second particle spring is attached to.
	 *  @param initRestLength rest length of spring.
	 *  @param initStiffness  stiffness/Hooke's constant.
	 *  @param initDamping    damping constant.
	 */
	public Spring(Particle initParticle1, Particle initParticle2, float initRestLength, float initStiffness, float initDamping)
	{
		particle1  = initParticle1;
		particle2  = initParticle2;
		damping    = initDamping;
		stiffness  = initStiffness;
		
		// If spring it at its rest length already
		if (initRestLength == AT_REST_LENGTH)
		{
			// Store current length
			restLength = new Vector2d(initParticle1.getPosition(), initParticle2.getPosition()).getLength();
		}
		else
		{
			restLength = initRestLength;
		}
	}

	/** Returns whether a particle is attached to this spring.
	 *  @param  particle particle to test.
	 *  @return if particle is attached to spring.
	 */
	public boolean isAttachedTo(Particle particle)
	{
		// Return true is particle is attached to spring
		return (particle == particle1 || particle == particle2);
	}
	
	/** Applies forces to attached particles. */
	public void update()
	{
		// Vector from first to second attached particle
		Vector2d relativePosition = new Vector2d(particle1.getPosition(), particle2.getPosition());
		// Relative velocity of particles vector
		Vector2d relativeVelocity = new Vector2d(particle1.getVelocity(), particle2.getVelocity());

		// Calculate spring length
		float springLength = relativePosition.getLength();
		// Calculate force spring will exert
		float springForce = -stiffness * (springLength - restLength);
		// Calculate spring damping force
		float dampingForce = (-damping * relativeVelocity.dot(relativePosition)) / springLength;
		// Calculate final force from spring
		Vector2d totalForce = relativePosition.normalise().multiply(springForce + dampingForce);

		// Apply force to second particle
		particle2.applyForce(totalForce);
		// Apply force to first particle in opposite direction
		particle1.applyForce(totalForce.reverse());
	}

	/** Returns first particle's position. */
	public Vector2d getParticle1Position()
	{
		return new Vector2d(particle1.getPosition());
	}

	/** Returns second particle's position. */
	public Vector2d getParticle2Position()
	{
		return new Vector2d(particle2.getPosition());
	}
}