// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.Dimension;

/** Represents a 2D vector. */
public class Vector2d
{
	/** Vector at origin i.e. (0, 0). */
	public static final Vector2d ORIGIN = new Vector2d(0, 0);
	
	/** Vector x-axis position. */
	public float x = 0;
	/** Vector y-axis position. */
	public float y = 0;

	/** Constructs a vector at the origin. */
	public Vector2d()
	{
	}

	/** Constructs a vector.
	 *  @param initX x-axis position.
	 *  @param initY y-axis position.
	 */
	public Vector2d(float initX, float initY)
	{
		x = initX;
		y = initY;
	}

	/** Constructs a vector that points from a start to an end position i.e. (end - start).
	 *  @param start start vector.
	 *  @param end   end vector.
	 */
	public Vector2d(Vector2d start, Vector2d end)
	{
		x = end.x - start.x;
		y = end.y - start.y;
	}

	/** Constructs a vector from another.
	 *  @param vector vector to clone.
	 */
	public Vector2d(Vector2d vector)
	{
		x = vector.x;
		y = vector.y;
	}

	/** Returns addition of two vectors.
	 *  @param  rhs vector to add.
	 *  @return     (this + rhs).
	 */
	public Vector2d add(Vector2d rhs)
	{
		return new Vector2d(x + rhs.x, y + rhs.y);
	}

	/** Returns subtraction of two vectors.
	 *  @param  rhs vector to subtract.
	 *  @return     (this - rhs).
	 */
	public Vector2d subtract(Vector2d rhs)
	{
		return new Vector2d(x - rhs.x, y - rhs.y);
	}

	/** Returns multiplication of two vectors.
	 *  @param  rhs vector to multiply by.
	 *  @return     (this * rhs).
	 */
	public Vector2d multiply(Vector2d rhs)
	{
		return new Vector2d(x * rhs.x, y * rhs.y);
	}

	/** Returns scaled vector.
	 *  @param  rhs scale factor.
	 *  @return     (this * rhs).
	 */
	public Vector2d multiply(float rhs)
	{
		return new Vector2d(x * rhs, y * rhs);
	}

	/** Returns division of two vector.
	 *  @param  rhs vector to divide by.
	 *  @return     (this / rhs).
	 */
	public Vector2d divide(Vector2d rhs)
	{
		return new Vector2d(x / rhs.x, y / rhs.y);
	}

	/** Returns scaled vector.
	 *  @param  rhs scale factor.
	 *  @return     (this / rhs).
	 */
	public Vector2d divide(float rhs)
	{
		return new Vector2d(x / rhs, y / rhs);
	}

	/** Returns vector with direction reversed.
	 *  @return (this * -1).
	 */
	public Vector2d reverse()
	{
		return new Vector2d(-x, -y);
	}

	/** Returns dot product of two vectors.
	 *  @return (this dot rhs).
	 */
	public float dot(Vector2d rhs)
	{
		return x * rhs.x + y * rhs.y;
	}

	/** Returns length of vector. */
	public float getLength()
	{
		return (float)Math.sqrt(x * x + y * y);
	}

	/** Returns squared length of vector. */
	public float getSquaredLength()
	{
		return x * x + y * y;
	}

	/** Returns normalised vector. */
	public Vector2d normalise()
	{
		float length = getLength();

		return (length == 0 ? ORIGIN : new Vector2d(x / length, y / length));
	}

	/** Returns vector rotated around origin.
	 *  @param zRotate amount to rotate around z-axis.
	 *  @return        rotated vector.
	 */
	public Vector2d rotate(float zRotate)
	{
		return new Vector2d((float)(x * Math.cos(zRotate) - y * Math.sin(zRotate)), (float)(x * Math.sin(zRotate) + y * Math.cos(zRotate)));
	}
	
	/** Returns vector clipped to boundary.
	 *  @param  bounds clipping boundary.
	 *  @return        clipped vector.
	 */
	public Vector2d clip(Dimension bounds)
	{
		return new Vector2d(Math.max(Math.min(x, bounds.width), 0), Math.max(Math.min(y, bounds.height), 0));
	}

	public String toString()
	{
		return "(" + x + ", " + y + ")";
	}
}