// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.Dimension;
import java.util.Vector;

/** Represents a world consisting of particles connected together with springs. Spring/particle indexes start at zero, are always consecutive and are unique. */
public class World
{
	/** Index representing no particle list or spring list index. */
	public static final int NONE = -1;

	/** List of particles in world. */
	private Vector particleList = new Vector();
	/** List of springs in world. */
	private Vector springList = new Vector();

	/** World boundary dimensions. */
	private Dimension  bounds;

	/** Gravitational pull vector. */
	private Vector2d gravitationalPull;
	/** World damping coeffecient. */
	private float damping;
	/** Particle coefficient of restitution. */
	private float coefficientOfRestitution;
	/** Speed where particles are considered not moving. */
	private float staticSpeed;
	/** Particle coefficent of static friction with boundaries. */
	private float staticFriction;
	/** Particle kinetic friction with boundaries. */
	private float kineticFriction;

	/** Constructs a new world.
	 *  @param initBounds                   boundaries of world.
	 *  @param initDamping                  damping constant to be applied to all particles
	 *  @param initCoefficientOfRestitution coefficient of restitution between particles and boundary.
	 *  @param initStaticSpeed              speed where particles are considered not moving.
	 *  @param initStaticFriction           static friction coefficent between particles and boundary.
	 *  @param initKineticFriction          kinetic friction coefficent between particles and boundary.
	 *  @param initGravitationalPull        vector multiplied by particle mass to get gravitational pull on that mass.
	 */
	public World(Dimension initBounds, float initDamping, float initCoefficientOfRestitution, float initStaticSpeed, float initStaticFriction, float initKineticFriction, Vector2d initGravitationalPull)
	{
		bounds                   = new Dimension(initBounds);
		damping                  = initDamping;
		coefficientOfRestitution = initCoefficientOfRestitution;
		staticSpeed              = initStaticSpeed;
		staticFriction           = initStaticFriction;
		kineticFriction          = initKineticFriction;
		gravitationalPull        = new Vector2d(initGravitationalPull);
	}

	/** Adds a new particle.
	 *  @param  particle particle to add.
	 *  @return          index of particle that was added.
	 */
	public int addParticle(Particle particle)
	{
		// Add particle
		particleList.addElement(particle);
		// Return particle's position in list
		return particleList.size() - 1;
	}

	/** Adds a new spring.
	 *  @param  spring spring to add.
	 *  @return        index of spring that was added.
	 */
	public int addSpring(Spring spring)
	{
		// Add spring
		springList.addElement(spring);		
		// Return spring's index in list
		return springList.size() - 1;
	}

	/** Returns number of particle in world. */
	public int getNumParticles()
	{
		return particleList.size();
	}

	/** Returns particle from index. */
	public Particle getParticle(int index)
	{
		return (Particle)particleList.elementAt(index);
	}

	/** Returns number of springs in world. */
	public int getNumSprings()
	{
		return springList.size();
	}
	
	/** Returns spring by index. */
	public Spring getSpring(int index)
	{
		return (Spring)springList.elementAt(index);
	}

	/** Removes particle from world along with any springs attached to it. Last added particle is moved into specified index.
	  * @param index index of particle.
	  */
	public void removeParticle(int index)
	{
		// Get particle
		Particle particle = getParticle(index);

		// For each spring
		for (int i = 0; i < getNumSprings(); ++i)
		{
			// If spring is attached to particle
			if (getSpring(i).isAttachedTo(particle))
			{
				// Remove current spring
				removeSpring(i);
				// Go back one spring to check one that replaced current
				--i;
			}
		}
		
		// Copy last element into particle's position
		particleList.setElementAt(getParticle(getNumParticles() - 1), index);
		// Shrink list by one
		particleList.setSize(getNumParticles() - 1);
	}

	/** Removes spring from world. Last added spring is moved into specified index.
	  * @param index index of spring.
	  */
	public void removeSpring(int index)
	{
		// Copy last element into particle's position
		springList.setElementAt(getSpring(getNumSprings() - 1), index);
		// Shrink list by one
		springList.setSize(getNumSprings() - 1);
	}

	/** Returns one particle at specified position.
	 *  @param position position to test at.
	 *  @param radius   radius particle can be within of <code>position</code>.
	 *  @return         index of particle if one was found, <code>NONE</code> if there was not.
	 */
	public int getParticle(Vector2d position, float radius)
	{
		// For each particle
		for (int i = 0; i < getNumParticles(); ++i)
		{
			// If particle is within specified radius of position
			if (new Vector2d(position, getParticle(i).getPosition()).getLength() <= radius)
			{
				// Return particle index
				return i;
			}
		}

		// No particle near position
		return NONE;
	}

	/** Returns if a spring is attached to two particles.
	 *  @param particleIndex1 first particle index.
	 *  @param particleIndex2 second particle index.
	 *  @return               whether they are attached by a spring.
	 */
	public boolean isSpringAttached(int particleIndex1, int particleIndex2)
	{
		// Get both particles
		Particle particle1 = getParticle(particleIndex1);
		Particle particle2 = getParticle(particleIndex2);

		// For each spring
		for (int i = 0; i < getNumSprings(); ++i)
		{
			// Current spring
			Spring spring = getSpring(i);

			// If both particles are attached to spring
			if (spring.isAttachedTo(particle1) && spring.isAttachedTo(particle2))
			{
				// Spring is attached between particles
				return true;
			}
		}
		
		// Spring is not attached between particles
		return false;
	}

	/** Updates all springs and particles in world.
	 *	@param deltaTime amount to advance world in seconds.
	 */
	public void update(float deltaTime)
	{
		// For each particle
		for (int i = 0; i < getNumParticles(); ++i)
		{
			// Apply gravitational pull
			getParticle(i).applyForce(gravitationalPull.multiply(getParticle(i).getMass()));
		}

		// For each spring
		for (int i = 0; i < getNumSprings(); ++i)
		{
			// Update spring forces
			getSpring(i).update();
		}

		// For each particle
		for (int i = 0; i < getNumParticles(); ++i)
		{
			// Apply collision forces, with friction
			getParticle(i).applyCollisions(true, bounds, coefficientOfRestitution, staticSpeed, staticFriction, kineticFriction);
		}

		// For each particle
		for (int i = 0; i < getNumParticles(); ++i)
		{
			// Apply damping
			getParticle(i).applyDamping(damping);
		}

		// For each particle
		for (int i = 0; i < getNumParticles(); ++i)
		{
			// Update particle position
			getParticle(i).update(deltaTime);
		}
	}
}